"""
The full documentation is avaiable here: https://docs.optimeering.no/

These are the functions available when importing __archimedes__.

Example usage:
>>> archimedes.list_ids()
...prints list of available time series
"""

from archimedes.auth import archimedes_auth
from archimedes.configuration import constants  # config
from archimedes.data.api import (
    get,
    get_intraday_trades,
    get_latest,
    list_ids,
    list_series_price_areas,
    get_predictions,
    list_prediction_ids,
    get_predictions_ref_dts,
    store_prediction,
    forecast_list_ref_times,
    forecast_diff,
    forecast_get_by_ref_time_interval,
)
from archimedes.organizer.organizer import (
    run,
    store,
    load_latest_model,
    log,
    deploy,
    store_test_results,
    add_tag,
    set_note,
    store_model,
)
from archimedes.utils import compact_print, full_print, environ


__all__ = [
    "archimedes_auth",
    "compact_print",
    "full_print",
    "environ",
    "get",
    "get_intraday_trades",
    "get_latest",
    "list_ids",
    "list_series_price_areas",
    "load_latest_model",
    "log",
    "run",
    "store",
    "store_test_results",
    "deploy",
    "add_tag",
    "set_note",
    "get_predictions",
    "get_predictions_ref_dts",
    "list_prediction_ids",
    "store_model",
    "store_prediction",
    "forecast_list_ref_times",
    "forecast_diff",
    "forecast_get_by_ref_time_interval",
]



import os
from typing import Optional
import abc
import pandas as pd
from dynaconf import Dynaconf
import requests

# Global archimedes config
USER_HOME_DIR = os.path.expanduser("~")
ARCHIMEDES_CONF_DIR = os.path.join(USER_HOME_DIR, ".archimedes")
SAVED_CONFIG_PATH = os.path.join(ARCHIMEDES_CONF_DIR, "arcl.json")
SAVED_MSAL_TOKEN_CACHE_PATH = os.path.join(ARCHIMEDES_CONF_DIR, "msal.cache.bin")


# Local archimedes config
CONFIG_FILE_NAME = "archimedes.toml"
CONFIG_ENV_VAR_PREFIX = "ARCHIMEDES"


class ConfigNotFoundException(Exception):
    pass


def _find_config_file_path(path=None) -> Optional[str]:
    """Find the path to the archimedes.toml file"""
    if path is None:
        path = os.path.abspath(os.getcwd())

    if CONFIG_FILE_NAME in os.listdir(path):
        return os.path.join(path, CONFIG_FILE_NAME)

    new_path = os.path.dirname(path)
    if new_path == path:
        return None

    return _find_config_file_path(new_path)


def _find_root_folder():
    """Find the root folder of the project"""
    config_file_path = _find_config_file_path()
    if config_file_path is None:
        raise FileNotFoundError(f"Missing the {CONFIG_FILE_NAME} configuration file.")
    return os.path.dirname(config_file_path)


def get_config():
    config_file_path = os.getenv("ARCHIMEDES_CONFIG_FILE_PATH")
    return Dynaconf(
        envvar_prefix=CONFIG_ENV_VAR_PREFIX,
        settings_files=[_find_config_file_path(config_file_path)],
    )


config = get_config()


class ArchimedesConstants:
    DATE_LOW = pd.to_datetime("1900-01-01T00:00:00+00:00")
    DATE_HIGH = pd.to_datetime("2090-01-01T00:00:00+00:00")


constants = ArchimedesConstants()


ARCHIMEDES_API_CONFIG_URL = "https://arcl.optimeering.no/config.json"


class InvalidEnvironmentException(Exception):
    pass


class ApiConfig(abc.ABC):
    def __init__(self, env):
        config_result = requests.get(ARCHIMEDES_API_CONFIG_URL)
        self.config = config_result.json()
        if env not in self.config.keys():
            raise InvalidEnvironmentException(
                f"Invalid environment {env}, "
                f"supported values are {', '.join([str(key) for key in self.config.keys()])}"
            )
        self.environment = env.lower()

    def __getattr__(self, item):
        env_config = self.config[self.environment]
        return env_config[item]


environment_override = os.getenv("ENVIRONMENT")
environment = environment_override if environment_override else config.project.environment

api_config = ApiConfig(environment)



import os

import msal

from .configuration import api_config
from .token_cache import token_cache

API_APP_ID_URI_BASE = f"api://{api_config.aad_app_client_id}"
SCOPES = [
    f"{API_APP_ID_URI_BASE}/.default",
]


class NoneAuth(Exception):
    """User not logged in. Please log in using `arcl auth login <organization_id>"""

    pass


class ArchimedesPublicClientAuth:
    def __init__(self, client_id, authority, cache=None):
        self.app = self.build_msal_app(client_id, authority, cache=cache)

    @staticmethod
    def build_msal_app(client_id, authority, cache=None):
        return msal.PublicClientApplication(
            client_id,
            authority=authority,
            token_cache=cache,
        )


class ArchimedesLocalAuth(ArchimedesPublicClientAuth):
    def __init__(self):
        super().__init__(api_config.client_id, api_config.authority, token_cache)

    def get_access_token_silent(self):
        # We now check the cache to see
        # whether we already have some accounts that the end user already used to sign in before.
        accounts = self.app.get_accounts()
        if not accounts:
            return None

        chosen = accounts[0]
        result = self.app.acquire_token_silent(SCOPES, account=chosen)

        if result is None or "access_token" not in result:
            description = f" Error details: {result['error_description']}" if "error_description" in result else ""
            raise NoneAuth(
                f"User not logged in. Please log in using `arcl auth login <organization_id>`.{description}"
            )

        return result.get("access_token")


class ArchimedesConfidentialAuth:
    def __init__(self, client_id, client_credential, authority):
        self.app = self.build_confidential_msal_app(client_id, client_credential, authority)

    def get_access_token_silent(self):
        result = self.app.acquire_token_for_client(SCOPES)

        if result is None or "access_token" not in result:
            description = f" Error details: {result['error_description']}" if "error_description" in result else ""
            raise NoneAuth(
                "Authentication failed. Please make sure that AZURE_AD_APP_ID, AZURE_AD_APP_CLIENT_CREDENTIAL and "
                f"AZURE_AD_TENANT_ID are properly configured.{description}"
            )

        return result.get("access_token")

    @staticmethod
    def build_confidential_msal_app(client_id, client_credential, authority):
        return msal.ConfidentialClientApplication(
            client_id=client_id,
            client_credential=client_credential,
            authority=authority
        )


archimedes_auth = None
use_app_authentication = os.getenv("USE_APP_AUTHENTICATION", "false").strip().lower() not in ["false", "f", "0"]
use_web_authentication = os.getenv("USE_WEB_AUTHENTICATION", "false").strip().lower() not in ["false", "f", "0"]
assert not (use_app_authentication and use_web_authentication), \
    "Only one of USE_APP_AUTHENTICATION or USE_WEB_AUTHENTICATION can be set to TRUE"
is_local = not use_app_authentication and not use_web_authentication
if is_local:
    archimedes_auth = ArchimedesLocalAuth()
else:
    azure_ad_tenant_id = os.getenv("AZURE_AD_TENANT_ID")
    azure_ad_client_id = os.getenv("AZURE_AD_APP_ID")
    azure_ad_authority = f"https://login.microsoftonline.com/{azure_ad_tenant_id}"

    if use_app_authentication:
        azure_ad_client_credential = os.getenv("AZURE_AD_APP_CLIENT_CREDENTIAL")
        archimedes_auth = ArchimedesConfidentialAuth(azure_ad_client_id, azure_ad_client_credential, azure_ad_authority)
    elif use_web_authentication:
        archimedes_auth = None



"""
A collection of functions for integrating with the Archimedes API
"""

import json
from http import HTTPStatus
from typing import Dict, List

import pandas as pd
import requests
from requests.exceptions import HTTPError, JSONDecodeError

import archimedes
from archimedes.auth import archimedes_auth, NoneAuth
from archimedes.configuration import api_config

API_URL = api_config.url
API_VERSION = 2


def get_intraday_trades(
    price_areas: List[str] = None,
    start: str = None,
    end: str = None,
    *,
    access_token: str = None,
):
    """Get raw intraday trades from Archimedes Database

    This function can be used to fetch raw time series from the Archimedes Database without any post-processing.
    To see which series are available, use `list_ids()`.

    Example:
        >>> archimedes.get(
        >>>     price_areas=["NO1",],
        >>>     start="2020-06-20T04:00:00+00:00",
        >>>     end="2020-06-20T09:00:00+00:00",
        >>> )
                from_dt to_dt   series_id   version price   trade_time  buy_area    sell_area   attributes
        0	2022-03-01T00:00:00+00:00	2022-03-01T01:00:00+00:00	NP/IntradayTrades	1646193645	146.63	2022-02-28T17:02:06.934+0000	OPX	NO1	{'price': 146.63,...,  'product_code': 'PH-20220301-01'}
        1	2022-03-01T00:00:00+00:00	2022-03-01T01:00:00+00:00	NP/IntradayTrades	1646193645	146.63	2022-02-28T17:02:06.934+0000	OPX	NO1	{'price': 146.63,... , 'product_code': 'PH-20220301-01'}
        2	2022-03-01T00:00:00+00:00	2022-03-01T01:00:00+00:00	NP/IntradayTrades	1646193645	146.63	2022-02-28T17:02:06.934+0000	OPX	NO1	{'price': 146.63, ..., 'product_code': 'PH-20220301-01'}
        ...
        155	2022-03-01T23:00:00+00:00	2022-03-02T00:00:00+00:00	NP/IntradayTrades	1646193645	148.99	2022-03-01T13:05:40.934+0000	NL	NO1	{'price': 148.99,... , 'product_code': 'PH-20220301-24'}
        156	2022-03-01T23:00:00+00:00	2022-03-02T00:00:00+00:00	NP/IntradayTrades	1646193645	148.8	2022-03-01T12:57:58.777+0000	NO1	FI	{'price': 148.8,...,  'product_code': 'PH-20220301-24'}
        157	2022-03-01T23:00:00+00:00	2022-03-02T00:00:00+00:00	NP/IntradayTrades	1646193645	148.8	2022-03-01T12:58:02.115+0000	NO1	FI	{'price': 148.8, ... , 'product_code': 'PH-20220301-24'}


    Args:
        price_areas (List[str], optional): The price areas to pick, all price areas if None. Defaults to None.
        start (str, optional): The first datetime to fetch (inclusive). Returns all if None. Defaults to None.
        end (str, optional): The last datetime to fetch (exclusive). Returns all if None. Defaults to None.
        access_token (str, optional): None - access token for the API

    Returns:
        DataFrame with all the time series data

    Raises:
        HTTPError: If an http error occurs when requesting the API.
        NoneAuth: If the user is unauthorized or if the authorization has expired.
    """

    if isinstance(price_areas, str):
        price_areas = [price_areas]

    if start is None:
        start = archimedes.constants.DATE_LOW
    else:
        start = pd.to_datetime(start)

    if end is None:
        end = archimedes.constants.DATE_HIGH
    else:
        end = pd.to_datetime(end)

    query = {
        "series_id": "NP/IntradayTrades",
        "price_areas": price_areas,
        "start": start,
        "end": end,
        "flatten_columns": False,
    }

    observation_data = _make_api_request(
        f"{API_URL}/v{API_VERSION}/observation_json/get",
        access_token=access_token,
        params=query,
    )

    if not len(observation_data):
        return pd.DataFrame(
            columns=[
                "from_dt",
                "to_dt",
                "series_id",
                "version",
                "price",
                "trade_time",
                "buy_area",
                "sell_area",
                "attributes",
            ]
        )
    observation_data = [
        {
            **i,
            "price": i["value"].get("price"),
            "trade_time": i["value"].get("trade_time"),
        }
        for i in observation_data
    ]
    observation_data = pd.DataFrame.from_dict(observation_data)

    # Extracting price area and series id
    observation_data[
        ["series_id1", "series_id2", "buy_area", "sell_area"]
    ] = observation_data["series_id"].str.split("/", 3, expand=True)
    observation_data["series_id"] = (
        observation_data["series_id1"] + "/" + observation_data["series_id2"]
    )
    observation_data["attributes"] = observation_data["value"]
    observation_data["attributes_str"] = observation_data["value"].astype(str)
    observation_data = observation_data.drop_duplicates(
        ["from_dt", "series_id", "attributes_str"]
    )
    observation_data = observation_data.drop(
        ["series_id1", "series_id2", "value", "version", "attributes_str"], axis=1
    )
    observation_data = observation_data.sort_values(by=["from_dt"]).reset_index(
        drop=True
    )

    return observation_data


def get(
    series_ids: List[str],
    price_areas: List[str] = None,
    start: str = None,
    end: str = None,
    *,
    access_token: str = None,
):
    """Get any number of time series.

    This function can be used to fetch time series from the Archimedes Database.
    To see which series are available, use `list_ids()`.

    Example:
        >>> archimedes.get(
        >>>     series_ids=["NP/AreaPrices"],
        >>>     price_areas=["NO1", "NO2"],
        >>>     start="2020-06-20T04:00:00+00:00",
        >>>     end="2020-06-28T04:00:00+00:00",
        >>> )
        series_id                 NP/AreaPrices
        price_area                          NO1   NO2
        from_dt
        2020-06-20T04:00:00+00:00          1.30  1.30
        2020-06-20T05:00:00+00:00          1.35  1.35
        ...                                 ...   ...
        2020-06-28T03:00:00+00:00          0.53  0.53
        2020-06-28T04:00:00+00:00          0.55  0.55

    Args:
        series_ids (List[str]): The series ids to get.
        price_areas (List[str], optional): The price areas to pick, all price areas if None. Defaults to None.
        start (str, optional): The first datetime to fetch (inclusive). Returns all if None. Defaults to None.
        end (str, optional): The last datetime to fetch (exclusive). Returns all if None. Defaults to None.
        access_token (str, optional): None - access token for the API

    Returns:
        DataFrame with all the time series data

    Raises:
        HTTPError: If an http error occurs when requesting the API.
        NoneAuth: If the user is unauthorized or if the authorization has expired.
    """

    if isinstance(series_ids, str):
        series_ids = [series_ids]

    if isinstance(price_areas, str):
        price_areas = [price_areas]

    if start is None:
        start = archimedes.constants.DATE_LOW
    else:
        start = pd.to_datetime(start)

    if end is None:
        end = archimedes.constants.DATE_HIGH
    else:
        end = pd.to_datetime(end)

    query = {
        "series_ids": series_ids,
        "price_areas": price_areas,
        "start": start,
        "end": end,
        "flatten_columns": True,
    }

    df = _make_api_request(
        f"{API_URL}/v{API_VERSION}/data/get",
        access_token=access_token,
        params=query,
    )

    if len(df):
        df = pd.DataFrame.from_dict(df)
        df = df.sort_values(by=["from_dt", "version"])
    else:
        df = pd.DataFrame(
            columns=["series_id", "from_dt", "ref_dt", "value", "version", "price_area"]
        )
    df = df.pivot_table(
        values="value",
        columns=["series_id", "price_area"],
        index="from_dt",
        aggfunc="last",
    )

    del query["series_ids"]
    for series_id in series_ids:
        query["series_id"] = series_id
        observation_data = _make_api_request(
            f"{API_URL}/v{API_VERSION}/observation_json/get",
            access_token=access_token,
            params=query,
        )
        if len(observation_data):
            observation_data = pd.DataFrame.from_dict(observation_data)

            observation_data[
                ["series_id1", "series_id2", "price_area"]
            ] = observation_data["series_id"].str.split("/", 2, expand=True)
            observation_data["series_id"] = (
                observation_data["series_id1"] + "/" + observation_data["series_id2"]
            )
            observation_data = observation_data.drop(
                ["series_id1", "series_id2"], axis=1
            )

            if series_id == "NP/IntradayTrades":

                def agg_func(x):
                    return list(x)

                sort_headers = ["from_dt"]
            else:
                agg_func = "last"
                sort_headers = ["from_dt", "version"]

            observation_data = observation_data.sort_values(by=sort_headers)
            observation_data = observation_data.pivot_table(
                values="value",
                columns=["series_id", "price_area"],
                index="from_dt",
                aggfunc=agg_func,
            )

            df = pd.merge(
                df, observation_data, left_index=True, right_index=True, how="outer"
            )

    df.index = pd.to_datetime(df.index)
    df = df.fillna("")
    return df


def get_latest(
    series_ids: List[str],
    price_areas: List[str] = None,
    *,
    access_token: str = None,
):
    """Get the most recent data for any number of time series.

    This function is similar to `get()`, but only fetches data from the past 48 hours,
    potentially including future hours as well (as in the case of Spot price data).

    @TODO: Add an argument `hours` that allows the 'lookback' period to be extended
    to an arbitrary number of hours.

    Example:
        >>> # Calling this function at 2020-03-15T10:15:00
        >>> archimedes.get_latest(
        >>>     series_ids=["NP/AreaPrices", "NP/ConsumptionImbalancePrices"],
        >>>     price_areas=["NO1"],
        >>> )
        series_id                 NP/AreaPrices  NP/ConsumptionImbalancePrices
        price_area                          NO1                            NO1
        from_dt
        2020-03-14T04:11:00+00:00          1.30                           1.30
        2020-03-14T05:12:00+00:00          1.35                           1.35
        ...                                 ...                            ...
        2020-03-15T22:00:00+00:00          0.53                            NaN
        2020-03-15T23:00:00+00:00          0.55                            NaN

    Args:
        series_ids (List[str]): The series ids to get.
        price_areas (List[str], optional): The price areas to pick, all price areas if None. Defaults to None.
        access_token (str, optional): None - access token for the API

    Returns:
        DataFrame with the latest time series data

    Raises:
        HTTPError: If an http error occurs when requesting the API.
        NoneAuth: If the user is unauthorized or if the authorization has expired.
    """
    now_dt = pd.Timestamp.now(tz="utc")
    start_dt = now_dt - pd.Timedelta(days=2)
    # +14 days should be enough in all cases now:
    end_dt = now_dt + pd.Timedelta(days=14)

    df = get(
        series_ids=series_ids,
        price_areas=price_areas,
        start=start_dt.isoformat(),
        end=end_dt.isoformat(),
        access_token=access_token,
    )

    return df


def get_predictions(
    series_ids: List[str] = None,
    price_areas: List[str] = None,
    start: str = None,
    end: str = None,
    ref_dt_start: str = None,
    ref_dt_end: str = None,
    *,
    access_token: str = None,
) -> pd.DataFrame:
    """Get any number of predictions

    This function can be used to fetch predictions from the Archimedes Database.

    Unlike `archimedes.get`, this will return a list, not a dataframe.

    Example:
        >>> archimedes.get_predictions(
            series_ids=["PX/rk-naive"],
            price_areas=["NO1"],
            start="2020"
        )
        >>> [...]

    Args:
        series_ids (List[str], optional): The series ids to get.
        price_areas (List[str], optional): The price areas to get the data for.
        start (str, optional):
            The first datetime to fetch (inclusive). Returns all if None. Defaults to None.
        end (str, optional):
            The last datetime to fetch (exclusive). Returns all if None. Defaults to None.
        ref_dt_start (pd.Timestamp, optional):
            The earliest ref_dt to fetch (inclusive). Defaults to None.
        ref_dt_end (pd.Timestamp, optional):
            The latest ref_dt to fetch (exclusive). Defaults to None.
        access_token (str, optional): None - access token for the API

    Returns:
        DataFrame with all the prediction data
    """
    if isinstance(series_ids, str):
        series_ids = [series_ids]

    if isinstance(price_areas, str):
        price_areas = [price_areas]

    if start is None:
        start = archimedes.constants.DATE_LOW
    else:
        start = pd.to_datetime(start)

    if end is None:
        end = archimedes.constants.DATE_HIGH
    else:
        end = pd.to_datetime(end)

    query = {"start": start, "end": end}

    if ref_dt_start is not None:
        query["ref_dt_start"] = pd.to_datetime(ref_dt_start)

    if ref_dt_end is not None:
        query["ref_dt_end"] = pd.to_datetime(ref_dt_end)

    if series_ids is not None:
        query["series_ids"] = series_ids

    if price_areas is not None:
        query["price_areas"] = price_areas

    data = _make_api_request(
        f"{API_URL}/v{API_VERSION}/data/get_predictions",
        access_token=access_token,
        params=query,
    )

    for item in data:
        json_data = item.get("json_data")
        item.update(json_data)
        del item["json_data"]

    df = pd.DataFrame.from_dict(data)

    df["from_dt"] = pd.to_datetime(df["from_dt"], utc=True)
    df["run_dt"] = pd.to_datetime(df["run_dt"], utc=True)
    df["ref_dt"] = pd.to_datetime(df["ref_dt"], utc=True)

    return df


def list_series_price_areas(series_id: str, *, access_token: str = None):
    """
    Retrieve all of the price_areas which are available for the specified data series

    Example:
        >>> archimedes.list_series_price_areas('NP/AreaPrices')
           price_areas
        0          DK1
        1          DK2
        ...        ...
        10         SE3
        11         SE4

    Returns:
        Dataframe with all available price areas for the specified series_id
    """
    query = {
        "series_id": series_id,
    }
    data = _make_api_request(
        f"{API_URL}/v{API_VERSION}/data/list_series_price_areas",
        access_token=access_token,
        params=query,
    )
    data = pd.DataFrame.from_dict(data)

    observation_data = _make_api_request(
        f"{API_URL}/v{API_VERSION}/observation_json/list_series_price_areas",
        access_token=access_token,
        params=query,
    )
    observation_data = pd.DataFrame.from_dict(observation_data)

    price_area_df = pd.concat([data, observation_data]).drop_duplicates()
    price_area_df = price_area_df.sort_values("price_areas").reset_index(drop=True)
    return price_area_df


def list_ids(sort: bool = False, *, access_token: str = None):
    """List all the series ids available.

    Example:
        >>> archimedes.list_ids()
                                    series_id
        0   NP/NegativeProductionImbalancePrices
        1                    NP/ProductionTotals
        ..                                   ...
        38                 NP/OrdinaryDownVolume
        39                    NP/SpecialUpVolume

    Args:
        sort (bool): False - return all series in one dataframe column, True - order dataframe by data-origin
        access_token (str, optional): None - access token for the API

    Returns:
        DataFrame with all available list_ids
    """
    data = _make_api_request(
        f"{API_URL}/v{API_VERSION}/data/list_ids", access_token=access_token
    )
    data = pd.DataFrame.from_dict(data)

    observation_data = _make_api_request(
        f"{API_URL}/v{API_VERSION}/observation_json/list_ids", access_token=access_token
    )
    observation_data = pd.DataFrame.from_dict(observation_data)

    series_df = pd.concat([data, observation_data]).drop_duplicates()
    series_df = series_df.sort_values(["series_id"]).reset_index(drop=True)
    if not sort:
        return series_df

    series_df["pre"] = series_df["series_id"].str.split("/", 1).str[0]
    series_df = pd.DataFrame.from_dict(
        series_df.groupby("pre")["series_id"].apply(list).to_dict(), orient="index"
    ).transpose()
    series_df = series_df[sorted(series_df.columns)]

    series_df = series_df.fillna("")
    return series_df.copy()


def list_prediction_ids(*, access_token: str = None):
    """List all the prediction series ids available.

    Example:
        >>> archimedes.list_prediction_ids()
                                     series_id
        0               PX/rk-nn-probabilities
        1   PX/rk-nn-direction-probabilities/U
        ..                                ...
        22                           PX/rk-901
        23                         PX/rk-naive
    """

    data = _make_api_request(
        f"{API_URL}/v{API_VERSION}/data/list_prediction_ids", access_token=access_token
    )

    return pd.DataFrame.from_dict(data)


def get_predictions_ref_dts(prediction_id: str = None, *, access_token: str = None):
    """Get which ref_dts are available.

    ref_dt == prediction_build_dt
    Users views in the database.

    Args:
        prediction_id (str): The series id to get the reference dts for. If None, get ref_dts for all prediction_ids.
        access_token (str, optional): Access token for the API

    Returns:
        DataFrame with all ref_dts
    """
    query = {}

    if prediction_id:
        query["prediction_id"] = prediction_id

    data = _make_api_request(
        f"{API_URL}/v{API_VERSION}/data/get_predictions_ref_dts",
        access_token=access_token,
        params=query,
    )

    return pd.DataFrame.from_dict(data)


def _make_api_request(url, method="GET", access_token=None, *args, **kwargs):
    if access_token is None:
        if archimedes_auth is None:
            raise NoneAuth(
                "access_token parameter must be passed when using USE_WEB_AUTHENTICATION"
            )
        access_token = archimedes_auth.get_access_token_silent()

    headers = kwargs.pop("headers", None)
    if headers is None:
        headers = {}

    headers.update({"Authorization": f"Bearer {access_token}"})

    r = requests.request(method, url, headers=headers, *args, **kwargs)

    if r.status_code not in [HTTPStatus.OK, HTTPStatus.CREATED]:
        try:
            response_json = r.json()
            if "message" in response_json:
                error_message = response_json.get("message")
            elif "detail" in response_json:
                error_message = response_json.get("detail")
            else:
                error_message = json.dumps(response_json)
        except JSONDecodeError:
            error_message = r.content
        raise HTTPError(f"API Error: {error_message}")

    return r.json()


def store_prediction(
    prediction_id: str,
    from_dt: pd.Timestamp,
    ref_dt: pd.Timestamp,
    run_dt: pd.Timestamp,
    data: Dict,
    *,
    access_token: str = None,
):
    """Store a prediction

    Example:
        >>> import archimedes
        >>> import pandas as pd
        >>> from dateutil.tz import gettz
        >>> tz_oslo = gettz("Europe/Oslo")
        >>> prediction_id = "test-prediction-id"
        >>> from_dt = pd.Timestamp("2021-04-11 23:47:16.854775807", tz=tz_oslo)
        >>> ref_dt = pd.Timestamp("2021-04-10 23:00:00.000000000", tz=tz_oslo)
        >>> run_dt = pd.Timestamp.now(tz=tz_oslo)
        >>> data = {"direction": "D", "probability": 0.8632089971077396, "hours_ahead": 1, "price_area": "NO1"}
        >>> archimedes.store_prediction(
        >>>     prediction_id=prediction_id,
        >>>     from_dt=from_dt,
        >>>     ref_dt=ref_dt,
        >>>     run_dt=run_dt,
        >>>     data=data
        >>> )
        True
    """
    payload = {
        "prediction_id": prediction_id,
        "from_dt": from_dt.isoformat(),
        "ref_dt": ref_dt.isoformat(),
        "run_dt": run_dt.isoformat(),
        "data": data,
    }

    ret = _make_api_request(
        f"{API_URL}/v{API_VERSION}/data/store_prediction",
        method="POST",
        data=json.dumps(payload),
        headers={"Content-Type": "application/json"},
        access_token=access_token,
    )

    return ret.get("success", False)


def forecast_list_ref_times(
    series_id: str,
    start: pd.Timestamp = None,
    end: pd.Timestamp = None,
    limit: int = None,
    *,
    access_token: str = None,
):
    """
    List all forecast reference times (the times that the forecast was generated)

    Args:
        series_id (str): The ID of the data series to find all the forecast reference times (the time the forecast was
            generated). Retrieve the complete list of series (both forecasts and observations) using the list_ids
            resource.
        start (pd.Timestamp, optional): The first datetime to fetch (inclusive). Returns all if not set. Should be
            specified in ISO 8601 format.
            (eg - '2021-11-29T06:00:00+00:00')
        end (pd.Timestamp, optional): The last datetime to fetch (exclusive). Returns all if not set. Should be
            specified in ISO 8601 format.
            (eg - '2021-11-30T06:00:00+00:00')
        limit (int, optional): Limit the output to a specific number of entries. No limit if not specified.
        access_token (str, optional): Access token for the API

    Example:
        >>> start = pd.Timestamp("2022-01-09T06:00:00+00:00")
        >>> end = pd.Timestamp("2022-01-10T06:00:00+00:00")
        >>> archimedes.forecast_list_ref_times('MET/forecast_air_temperature_2m', start, end)
                                   ref_times
        0   2022-01-10T05:00:00.000000+00:00
        1   2022-01-10T04:00:00.000000+00:00
        ...                              ...
        22  2022-01-09T07:00:00.000000+00:00
        23  2022-01-09T06:00:00.000000+00:00

    Returns:
        Dataframe with all of the forecast reference times
    """
    query = {
        "series_id": series_id,
        "start": start.isoformat() if start is not None else None,
        "end": end.isoformat() if end is not None else None,
        "limit": limit,
    }

    data = _make_api_request(
        f"{API_URL}/v{API_VERSION}/forecast/list_ref_times",
        access_token=access_token,
        params=query,
    )

    return pd.DataFrame.from_dict(data)


def forecast_diff(
    comparison_type: str,
    series_ids: List[str],
    price_areas: List[str] = None,
    ref_time1: pd.Timestamp = None,
    ref_time2: pd.Timestamp = None,
    *,
    access_token: str = None,
):
    """
    Get the difference between two different forecasts.

    Args:
        comparison_type: The type of comparison to do to the two forecasts:
            forecast_update - how has the the forecast for a specific time range been updated. The two forecast
                reference times must be within ~60 hours of each other. Otherwise the output will be empty (because
                the forecasts don't overlap).

            forecast_diff - compares the forecasts of any two dates to indicate how different they are.
        series_ids: The ID of the data series to get (eg - 'MET/forecast_air_temperature_2m' or
            'MET/forecast_wind_speed_10m'). To specify multiple data series, include the series_ids parameter multiple
            times in the url.
            Retrieve the complete list of series using the list_ids resource.
        price_areas: The name of the price area(eg - 'NO2', 'NO5', or 'DE1-NO1'). To specify multiple price areas,
            include the price_areas parameter multiple times in the url.
            Retrieve the complete list of price areas available for a specified series ID using list_series_price_areas
            resource.
        ref_time1: Specify one of the two timestamps for when a forecast was created. Should be specified in ISO 8601
            format (eg - '2021-11-29T06:00:00+00:00').
        ref_time2: Specify another of the two timestamps for when a forecast was created. Should be specified in ISO
            8601 format (eg - '2021-11-29T06:00:00+00:00').
        access_token (str, optional): Access token for the API

    Returns:
        Dataframe with the diff of the two forecasts
    """
    assert comparison_type in [
        "forecast_update",
        "forecast_diff",
    ], f"Unknown comparison_type '{comparison_type}' (should be either 'forecast_update' or 'forecast_diff'"
    query = {
        "forecast_comparison_type": comparison_type,
        "series_ids": series_ids,
        "price_areas": price_areas,
        "ref_time1": ref_time1,
        "ref_time2": ref_time2,
    }

    data = _make_api_request(
        f"{API_URL}/v{API_VERSION}/forecast/diff",
        access_token=access_token,
        params=query,
    )

    return pd.DataFrame.from_dict(data)


def forecast_get_by_ref_time_interval(
    series_id: str,
    price_area: str,
    start: pd.Timestamp = None,
    end: pd.Timestamp = None,
    forecast_interval: int = 24,
    day_ahead_hour: int = None,
    *,
    access_token: str = None,
):
    """
    Get a single forecast value for every hour. The value should be from the forecast that was generated at least
    forecast_interval hours prior.

    Args:
        series_id: The ID of the data series to get (eg - 'MET/forecast_air_temperature_2m' or
            'MET/forecast_wind_speed_10m'). Retrieve the complete list of series using the list_ids resource.
        price_area: The name of the price area(eg - 'NO2', 'SE3'). Retrieve the complete list of price areas available
            for a specified series ID using list_series_price_areas resource.
        start: The first datetime to fetch (inclusive). Returns all if not set. Should be specified in ISO 8601 format
            (eg - '2021-11-29T06:00:00+00:00')
        end: The last datetime to fetch (exclusive). Returns all if not set. Should be specified in ISO 8601 format
            (eg - '2021-11-30T06:00:00+00:00')
        forecast_interval: The number of hours earlier that the forecast must have been generated. In some cases, it
            could be older (if no forecast was generated at exactly that hour). NOTE - this is ignored if
            day_ahead_hour is set.
        day_ahead_hour: Used for day-ahead market. Indicates the hour of the day when the market closes (CET - Central
            European Time). Will return the forecast generated before this time on the previous day
            ('forecast_interval' will be set to 24). For example, if set to '12' (noon CET), the values shown for every
            hour of a specific day will be fetched from the most recent forecast generated before noon (most likely
            11am) on the previous day.
        access_token (str, optional): Access token for the API

    Returns:
        Dataframe with the forecasted values
    """
    query = {
        "series_id": series_id,
        "price_area": price_area,
        "start": start,
        "end": end,
        "forecast_interval": forecast_interval,
        "day_ahead_hour": day_ahead_hour,
    }

    data = _make_api_request(
        f"{API_URL}/v{API_VERSION}/forecast/get_by_ref_time_interval",
        access_token=access_token,
        params=query,
    )

    return pd.DataFrame.from_dict(data)



import datetime
import json
from typing import List, Optional, Dict

from sqlalchemy.exc import ResourceClosedError

import archimedes
import pandas as pd
import numpy as np
import records
from archimedes.configuration import config


try:
    db_url = (
        f"postgresql://"
        f"{config.db.postgres_user}:"
        f"{config.db.postgres_pass}@"
        f"{config.db.postgres_host}:"
        f"{config.db.postgres_port}/"
        f"{config.db.postgres_dbname}"
    )
    db = records.Database(db_url=db_url)
except AttributeError as e:
    db = None
    db_error_msg = e


def list_ids(sort: bool = False):
    """List all the series ids available.

    Example:
        >>> archimedes.list_ids()
                                    series_id
        0   NP/NegativeProductionImbalancePrices
        1                      SN/FRRADownVolume
        ..                                   ...
        38                 NP/OrdinaryDownVolume
        39                    NP/SpecialUpVolume
    
    Args:
        sort (bool): False - return all series in one dataframe column, True - order dataframe by data-origin
    
    Returns:
        DataFrame with all available list_ids
    """
    if db is None:
        raise ValueError(db_error_msg)
    query = """
            SELECT distinct series_id from nordpool
            UNION
            SELECT distinct series_id from statnett
            """
    rows = db.query(query)
    ret = rows.export("df")

    if not sort:
        return ret
    
    # get list of data-sources and sort alphabetically
    data_sources = np.unique([ids.split("/")[0] for ids in ret.values.flatten()])
    data_sources = np.sort(data_sources)

    sorted_df = pd.DataFrame()
    for source in data_sources:
        source_ids = ret[pd.Series(ret.values.flatten()).str.startswith(source)]
        # sort series_ids alphabetically and reset dataframe index
        source_ids = source_ids.sort_values(by="series_id")
        source_ids = source_ids.reset_index(drop=True)
        # append column of series_ids to frame
        sorted_df = pd.concat([sorted_df, source_ids], axis=1)

    # set name of columns and replace nans with empty strings
    sorted_df.columns = data_sources
    sorted_df = sorted_df.fillna("")
    return sorted_df.copy()


def list_prediction_ids(composite_ids=False):
    """List all the prediction series ids available.

    Example:
        >>> archimedes.list_prediction_ids()
                                    series_id
        0   NP/NegativeProductionImbalancePrices
        1                      SN/FRRADownVolume
        ..                                   ...
        38                 NP/OrdinaryDownVolume
        39                    NP/SpecialUpVolume
    """
    if db is None:
        raise ValueError(db_error_msg)
    if composite_ids:
        query = """
        SELECT distinct composite_id from predictions
        """
    else:
        query = """
        SELECT distinct series_id from predictions
        """
    rows = db.query(query)
    return rows.export("df")


def get(
    series_ids: List[str],
    price_areas: List[str] = None,
    start: str = None,
    end: str = None,
    flatten_columns: bool = False,
    long_format: bool = False,
):
    """Get any number of time series.

    This function can be used to fetch time series from the Archimedes Database.
    To see which series are available, use `list_ids()`.

    Example:
        >>> archimedes.get(
        >>>     series_ids=["NP/AreaPrices"],
        >>>     price_areas=["NO1", "NO2"],
        >>>     start="2020-06-20T04:00:00+00:00",
        >>>     end="2020-06-28T04:00:00+00:00",
        >>> )
        series_id                 NP/AreaPrices
        price_area                          NO1   NO2
        from_dt
        2020-06-20T04:00:00+00:00          1.30  1.30
        2020-06-20T05:00:00+00:00          1.35  1.35
        ...                                 ...   ...
        2020-06-28T03:00:00+00:00          0.53  0.53
        2020-06-28T04:00:00+00:00          0.55  0.55

    Args:
        series_ids (List[str]): The series ids to get.
        price_areas (List[str], optional): The price areas to pick, all price areas if None. Defaults to None.
        start (str, optional): The first datetime to fetch (inclusive). Returns all if None. Defaults to None.
        end (str, optional): The last datetime to fetch (exclusive). Returns all if None. Defaults to None.
        flatten_columns (bool, optional): The column names are flattened if True. Defaults to False.
        long_format (str, optional): Should the dataframe be long (instead of wide). Defaults to False.

    Returns:
        DataFrame with all the time series data
    """
    if db == None:
        raise ValueError(db_error_msg)

    if isinstance(series_ids, str):
        series_ids = [series_ids]

    if isinstance(price_areas, str):
        price_areas = [price_areas]

    if start == None:
        start = archimedes.constants.DATE_LOW
    else:
        start = pd.to_datetime(start)

    if end == None:
        end = archimedes.constants.DATE_HIGH
    else:
        end = pd.to_datetime(end)

    query = """
    SELECT c.series_id, c.from_dt, c.price_area, c.value, c.version FROM (
        SELECT * FROM nordpool
        UNION
        SELECT * FROM statnett
    ) as c
    WHERE c.series_id IN :series_ids
    AND c.from_dt >= :start
    AND c.from_dt < :end
    """

    if price_areas is None:
        rows = db.query(
            query=query,
            series_ids=tuple(series_ids),
            start=start,
            end=end,
        )
    else:
        query = f"{query} AND c.price_area IN :price_areas"
        rows = db.query(
            query=query,
            series_ids=tuple(series_ids),
            price_areas=tuple(price_areas),
            start=start,
            end=end,
        )

    df = rows.export("df")

    row_count = df.shape[0]
    if row_count == 0:
        return df

    df["from_dt"] = pd.to_datetime(df["from_dt"])
    df = df.sort_values(by=["from_dt", "version"])

    if long_format:
        df = (
            df.groupby(["series_id", "from_dt", "price_area"]).agg("last").reset_index()
        )
        df.drop(["version"], axis=1, inplace=True)
        return df

    df = df.pivot_table(
        values="value",
        columns=["series_id", "price_area"],
        index="from_dt",
        aggfunc="last",
    )
    if flatten_columns:
        new_columns = ["/".join(list(column)) for column in df.columns]
        df.columns = new_columns
    df = df.astype(float)
    return df


def get_latest(
    series_ids: List[str],
    price_areas: List[str] = None,
    flatten_columns: bool = False,
):
    """Get the most recent data for any number of time series.

    This function is similar to `get()`, but only fetches data from the past 48 hours,
    potentially including future hours as well (as in the case of Spot price data).

    @TODO: Add an argument `hours` that allows the 'lookback' period to be extended
    to an arbitrary number of hours.

    Example:
        >>> # Calling this function at 2020-03-15T10:15:00
        >>> archimedes.get_latest(
        >>>     series_ids=["NP/AreaPrices", "NP/ConsumptionImbalancePrices"],
        >>>     price_areas=["NO1"],
        >>> )
        series_id                 NP/AreaPrices  NP/ConsumptionImbalancePrices
        price_area                          NO1                            NO1
        from_dt
        2020-03-14T04:11:00+00:00          1.30                           1.30
        2020-03-14T05:12:00+00:00          1.35                           1.35
        ...                                 ...                            ...
        2020-03-15T22:00:00+00:00          0.53                            NaN
        2020-03-15T23:00:00+00:00          0.55                            NaN

    Args:
        series_ids (List[str]): The series ids to get.
        price_areas (List[str], optional): The price areas to pick, all price areas if None. Defaults to None.
        flatten_columns (bool, optional): The column names are flattened if True. Defaults to False.

    Returns:
        DataFrame with all the time series data
    """
    now_dt = pd.Timestamp.now(tz="utc")
    start_dt = now_dt - datetime.timedelta(days=2)
    # +14 days should be enough in all cases now:
    end_dt = now_dt + datetime.timedelta(days=14)

    df = get(
        series_ids=series_ids,
        price_areas=price_areas,
        start=start_dt.isoformat(),
        end=end_dt.isoformat(),
        flatten_columns=flatten_columns,
    )

    return df


def get_predictions(
    series_ids: List[str],
    start: str = None,
    end: str = None,
    ref_dt: str = None,
    price_area: str = None,
) -> List:
    """Get any number of predictions

    This function can be used to fetch predictions from the Archimedes Database.

    Unlike `archimedes.get`, this will return a list, not a dataframe.

    @TODO: It could be that this function should also return a pd.DataFrame,
    where the user can choose whether to have the 'wide' or 'long' format returned.

    Example:
        >>> archimedes.get_predictions(
            series_ids=["PX/rk-naive"],
            start="2020"
        )
        >>> [...]

    Args:
        series_ids (List[str]): The series ids to get.
        start (str, optional):
            The first datetime to fetch (inclusive). Returns all if None. Defaults to None.
        end (str, optional):
            The last datetime to fetch (exclusive). Returns all if None. Defaults to None.
        ref_dt (str, optional):
            The reference datetime, eg. what was the time of the latest RK price?
        price_area (str, optional):
            The price area, eg. "NO2"

    Returns:
        List with all the prediction data
    """
    if db == None:
        raise ValueError(db_error_msg)

    if isinstance(series_ids, str):
        series_ids = [series_ids]

    if start == None:
        start = archimedes.constants.DATE_LOW
    else:
        start = pd.to_datetime(start)

    if end == None:
        end = archimedes.constants.DATE_HIGH
    else:
        end = pd.to_datetime(end)

    query = """
    SELECT
        from_dt,
        version,
        value,
        attributes->>'ref_dt' as ref_dt,
        attributes->>'model_version' as model_version,
        attributes->>'hours_ahead' as hours_ahead,
        attributes->>'price' as price,
        attributes->>'price_area' as price_area
    FROM predictions as c
    WHERE c.series_id IN :series_ids
    AND c.from_dt >= :start
    AND c.from_dt < :end
    """

    params = {"series_ids": tuple(series_ids), "start": start, "end": end}

    if ref_dt != None:
        ref_dt = pd.to_datetime(ref_dt).isoformat()
        query += " AND c.attributes->>'ref_dt' = :ref_dt"
        params["ref_dt"] = ref_dt

    if price_area != None:
        query += " AND c.attributes->>'price_area' = :price_area"
        params["price_area"] = price_area

    rows = db.query(query, **params)

    data = json.loads(rows.export("json"))

    for d in data:
        d["from_dt"] = pd.to_datetime(d["from_dt"])
        d["ref_dt"] = pd.to_datetime(d["ref_dt"])
        d["hours_ahead"] = pd.to_timedelta(d["hours_ahead"])
        d["price"] = float(d["price"]) if d["price"] else None
        d["value"] = float(d["value"])

    return data


def get_predictions_ref_dts():
    """Get which ref_dts are available.

    ref_dt == prediction_build_dt
    Users views in the database.

    Returns:
        List[Dict]
    """
    if db == None:
        raise ValueError(db_error_msg)
    import time

    start = time.time()
    query = """
    SELECT * FROM v_predictions_ref_dts
    LIMIT 10
    """
    rows = db.query(query)
    end = time.time()
    print(f"database.py->get_predictions_ref_dts timing: {end - start}")
    # return_value = [record.as_dict() for record in rows.all()]
    return_value = json.loads(rows.export("json"))
    # from pprint import pprint
    # pprint(return_value)
    return return_value


def _get_predictions_UNRELEASED(
    series_ids: List[str],
    start: str = None,
    end: str = None,
    long_format: bool = True,
    rename_columns: Dict = None,
) -> pd.DataFrame:  # A replacement for get_predictions
    # Include create_composite_id in imports
    """Get any number of predictions

    This function can be used to fetch predictions from the Archimedes Database.

    Unlike `archimedes.get`, this will return a list, not a dataframe.

    @TODO: It could be that this function should also return a pd.DataFrame,
    where the user can choose whether to have the 'wide' or 'long' format returned.

    Example:
        >>> archimedes.get_predictions(
            series_ids=["PX/rk-naive"],
            start="2020"
        )
        >>> [...]

    Args:
        series_ids (List[str]): The series ids to get.
        start (str, optional):
            The first datetime to fetch (inclusive). Returns all if None. Defaults to None.
        end (str, optional):
            The last datetime to fetch (exclusive). Returns all if None. Defaults to None.
        long_format (str, optional):
            Should the dataframe be long (instead of wide). Defaults to True.

    Returns:
        Dataframe with all the prediction data
    """
    if db == None:
        raise ValueError(db_error_msg)

    if isinstance(series_ids, str):
        series_ids = [series_ids]

    if start == None:
        start = archimedes.constants.DATE_LOW
    else:
        start = pd.to_datetime(start)

    if end == None:
        end = archimedes.constants.DATE_HIGH
    else:
        end = pd.to_datetime(end)

    query = """
    SELECT * FROM predictions as c
    WHERE c.series_id IN :series_ids
    AND c.from_dt >= :start
    AND c.from_dt < :end
    """
    rows = db.query(query, series_ids=tuple(series_ids), start=start, end=end)

    df = rows.export("df")

    row_count = df.shape[0]
    if row_count == 0:
        return df

    df["from_dt"] = pd.to_datetime(df["from_dt"])
    df.drop(["created_at", "composite_id", "to_dt"], axis=1, inplace=True)
    attributes_df = pd.json_normalize(df["attributes"])
    df.drop(["attributes"], axis=1, inplace=True)
    for column in attributes_df.columns:
        if column in ["from_dt", "model_name", "value_description"]:
            pass
        elif column in ["hours_ahead"]:
            df[column] = pd.to_timedelta(attributes_df[column])
        elif column in ["ref_dt"]:
            df[column] = pd.to_datetime(attributes_df[column])
        elif column in ["price_area", "model_version"]:
            df[column] = attributes_df[column]
        elif column in ["quantile"]:
            df[column] = attributes_df[column]
            # df[column] = attributes_df[column].fillna("exp")
            # df[column] = df[column].apply(lambda x: x if x == "exp" else f"{int(x)}%")
        else:
            print("there was an error")
            print(column)
            raise ValueError("Uncaught column in attributes named: " + column)
        # TODO: add column using create_composite_id

    # TODO: return different format if format="wide"

    if rename_columns != None:
        for key, value in rename_columns.items():
            df[value] = df[key]
            df.drop([key], axis=1, inplace=True)

    # lst = [record.as_dict() for record in rows.all()]
    # return lst
    return df


def store_prediction(
    series_id: str,
    value: float,
    version: int,
    from_dt: pd.Timestamp,
    attributes: Optional[Dict] = None,
    composite_id: str = None,
):
    """Store a prediction

    Example:
        >>> # @TODO: Add example.

    @TODO:
        Include 'model' and / or 'model version' as arguments.

    Args:
        project_name (str): The name of the project
        model_name (str): The name of the model used to make the prediction
        target_dt (pd.Timestamp): The hour of the prediction
        last_data_dt (pd.Timestamp): The hour of the last available data used to make the prediction
        predicted_value (float): The value that has been predicted
        prediction_dt (pd.Timestamp, optional): The time at which the prediction was made
        target_series_id (str, optional): The Archimedes series id corresponding to the predicted value
    """
    if db == None:
        raise ValueError(db_error_msg)

    if attributes == None:
        attributes = {}

    for k, v in attributes.items():
        if isinstance(v, (pd.Timestamp, pd.Timedelta)):
            attributes[k] = v.isoformat()

    query = """
    INSERT INTO predictions (from_dt, value, version, attributes, series_id, composite_id)
    VALUES (:from_dt, :value, :version, :attributes, :series_id, :composite_id)
    """

    try:
        db.query(
            query,
            from_dt=from_dt,
            value=value,
            version=version,
            attributes=json.dumps(attributes),
            series_id=series_id,
            composite_id=composite_id,
        )
    except ResourceClosedError:
        # This is expected for records library when inserting
        pass
    return True



import copy
import datetime
import json
import logging
import os
import pathlib
import tempfile
import types
from typing import Callable, Dict, Union

import matplotlib
import matplotlib.pyplot as plt
import mlflow
import pandas as pd
from archimedes.configuration import config
from archimedes.utils.environ_context import environ

"""
Any code related to organizing models in MLFlow goes here.
"""

MLFLOW_ARTIFACT_CONFIG = {
    "AWS_ACCESS_KEY_ID": config.mlflow.artifacts_aws_access_key_id,
    "AWS_SECRET_ACCESS_KEY": config.mlflow.artifacts_aws_secret_access_key,
    "MLFLOW_S3_ENDPOINT_URL": config.mlflow.artifacts_s3_endpoint_url,
}


def mlflow_log_artifacts(*args, **kwargs):
    """Call mlflow.log_artifact by temporarily setting environment variables"""
    with environ(MLFLOW_ARTIFACT_CONFIG):
        mlflow.log_artifact(*args, **kwargs)


# Move flatten to utils
def _flatten(df):
    """Flatten the columns in one of our returned dataframes"""
    df_ = copy.deepcopy(df)
    new_columns = ["/".join(list(column)) for column in df_.columns]
    df_.columns = new_columns
    return df_


def _to_long_form(df):
    """Long form one of our returned dataframes"""
    # check if multi index or not, assuming not for now
    df_ = copy.deepcopy(df)
    df_ = _flatten(df_)
    df_.columns = ["value" + c for c in df_.columns]
    df_.columns = [c.replace("/", "_") for c in df_.columns]
    df_ = df_.reset_index()
    l = pd.wide_to_long(
        df_, stubnames="value", i="from_dt", j="series_id", suffix=r"\w+"
    )
    l = l.reset_index()
    return l


def _get_project_name():
    """Get project name from the configuration"""
    project_name = config.project.name
    return project_name


def _get_author():
    """Get author from the configuration"""
    author = config.project.author
    return author


def _get_environment():
    """Get environment from the configuration"""
    environment = config.project.environment
    return environment


# def _get_mlflow_tracking_uri():
#     """Get the MLflow Tracking URI from the configuration"""
#     config = _get_config()
#     mlflow_tracking_uri = config["mlflow"]["tracking_uri"]
#     return mlflow_tracking_uri


def _setup(local_mlflow: bool = False, autolog: bool = True):
    """Setup MLflow for our run

    Args:
        local_mlflow (bool, optional): If True, uses the local MLFlow. Defaults to False.
        autolog (bool, optional): Log extra data from the models. Defaults to True.

    Returns:
        dict: The context of the run
    """
    logging.basicConfig(level=logging.INFO, format="%(levelname)s:\t%(message)s")
    # tracking_uri = os.path.join(_find_root_folder(), "mlruns")
    # mlflow.set_tracking_uri(tracking_uri)
    if local_mlflow:
        tracking_uri = pathlib.Path(
            os.path.join(os.path.expanduser("~"), "mlruns")
        ).as_uri()
    else:
        tracking_uri = config.mlflow.tracking_url
    mlflow.set_tracking_uri(tracking_uri)

    if autolog:
        # As of MLflow v1.12.0, universal auto-logging is supported
        mlflow.autolog()

    return {
        "project_name": _get_project_name(),
        "author": _get_author(),
        "environment": _get_environment(),
    }


def log(message: str):
    """Log a message

    Args:
        message (str): The message to log
    """
    print(message)
    logging.info(message)


def run(
    func: Union[Callable, str],
    model_name: str,
    local_mlflow: bool = False,
    args: Dict = None,
):
    """Run a function, without deploying it.

    The first argument can be either a function, the path to a python file or a
    string on the format app:myfunction.

    Example:
        >>> def myfunction():
        >>>     x = 2
        >>>     print(f"The number x is {x}")
        >>> archimedes.run(myfunction, "My first function")
        INFO:	Starting run at 2020-08-20T23:03:53.788115
        INFO:	MLFlow URI: /Users/jo/mlruns
        hello
        INFO:	Ending run at 2020-08-20T23:03:53.794075
        INFO:	The run took 0:00:00.005960

    Args:
        func (Union[Callable, str]): The function to deploy.
        model_name (str): The name of the model you're running.
        local_mlflow (bool, optional): If True, uses the local MLFlow. Defaults to False.
        args (dict, optional): Arguments to pass to the function being run, @TODO: change
    """
    context = _setup(local_mlflow)
    mlflow.set_experiment(context["project_name"])
    mlflow.start_run(
        run_name=model_name,
    )
    mlflow.set_tags(context)
    mlflow.set_tag("run_type", "MANUAL")
    mlflow.set_tag("model_name", model_name)
    run_start = datetime.datetime.utcnow()
    logging.info("Starting run at %s" % run_start.isoformat())
    log("MLFlow URI: %s" % mlflow.get_tracking_uri())
    with environ(MLFLOW_ARTIFACT_CONFIG):
        if args != None:
            func(**args)
        else:
            func()
    mlflow.end_run()
    run_end = datetime.datetime.utcnow()
    run_delta = run_end - run_start
    logging.info("Ending run at %s" % run_end.isoformat())
    logging.info("The run took %s" % run_delta)


def deploy(func: Union[Callable, str], model_name: str, cron: str = None):
    """Deploy a function or script to run on the server

    This is the server equivalent of `archimedes.run`.

    Example:
        >>> my_function = lambda x: x+2
        >>> deploy(my_function, "My First Function", "*10***")

    Args:
        func (Union[Callable, str]): The function to deploy
        model_name (str): The name of the model
        cron (str, optional): A cron string. Defaults to None.

    Returns:
        str: The id of the deployed function
    """
    _configure_prefect_server_endpoint()
    from prefect import Flow, client, task
    from prefect.storage import Docker
    from prefect.executors import DaskExecutor
    from prefect.run_configs import KubernetesRun
    from prefect.schedules import Schedule
    from prefect.schedules.clocks import CronClock

    if cron:
        schedule = Schedule(clocks=[CronClock(cron)])
    else:
        schedule = None

    context = _setup()

    project_name = context["project_name"]
    prefect_client = client.Client(api_server=config.prefect.api_server)
    _create_prefect_project_if_not_exist(prefect_client, project_name)

    # We only deal with functions for now
    if isinstance(func, types.FunctionType):

        def wrapper_func():
            return run(func, model_name, local_mlflow=False)

        only_task = task(wrapper_func, name=model_name)  # name=model_to_run.__name__)

        flow = Flow(name=model_name, tasks=[only_task], schedule=schedule)

        flow.storage = Docker(
            registry_url=config.prefect.docker_registry, dockerfile="Dockerfile"
        )

        flow.run_config = KubernetesRun()
        flow.executor = DaskExecutor()

        print(flow)

        flow.register(project_name=project_name)

    elif ".py" in func:
        raise ValueError("Files not yet implemented")
    elif ":" in func:
        raise ValueError("Function string not yet implemented")
    else:
        raise ValueError("Not yet implemented.")


def _configure_prefect_server_endpoint():
    """Configure prefect to use server endpoint from archimedes configuration

    If prefect server environment is configured in archimedes configuration, it doesn't need to be read from the
    prefect configuration file .prefect/config.yaml
    """
    if config.prefect.server_endpoint:
        os.environ["PREFECT__BACKEND"] = "server"
        os.environ["PREFECT__SERVER__ENDPOINT"] = config.prefect.server_endpoint


def _create_prefect_project_if_not_exist(prefect_client, project_name):
    """Create prefect project if it doesn't already exist

    project_name is mandatory in prefect version 0.13.0 and higher when registering a flow. So, it needs to be created.
    """
    from prefect.utilities.graphql import with_args

    query_project = {
        "query": {
            with_args("project", {"where": {"name": {"_eq": project_name}}}): {
                "id": True
            }
        }
    }
    project = prefect_client.graphql(query_project).data.project
    if not project:
        prefect_client.create_project(project_name)


def _store_dataframe(df: pd.DataFrame, name: str):
    """Store a dataframe in mlflow

    Example usage:
        df = pd.DataFrame(...)
        _store_dataframe(df, "MySpecialDataFrame")
        >> # The dataframe is stored and stored along with the
        >> # other run data.

    Args:
        df: The dataframe to store
        name: The name to associate with the dataframe
    """
    # name = name.replace("/", "__")
    with tempfile.TemporaryDirectory() as temporary_directory:
        # path = os.path.join(temporary_directory, str(t.timestamp()) + ".csv")
        filename_df = name + ".csv"
        filename_fig = name + ".jpeg"
        path_df = os.path.join(temporary_directory, filename_df)
        path_figure = os.path.join(temporary_directory, filename_fig)

        flat_df = _flatten(df)
        flat_df.index = pd.to_datetime(flat_df.index)

        num_cols = len(flat_df.columns)
        axs = flat_df.plot(
            subplots=True,
            linewidth=0.5,
            layout=(num_cols, 1),
            figsize=(10, 2 * num_cols),
            sharex=True,
            sharey=False,
        )
        # plt.show()
        # print(type(axs))  # <class 'numpy.ndarray'>
        # print(type(axs[0]))  # <class 'numpy.ndarray'>
        # print(type(axs[0][0]))  # <class 'matplotlib.axes._subplots.AxesSubplot'>
        fig = axs[0][0].get_figure()
        fig.savefig(path_figure)

        df.to_csv(path_df)
        mlflow_log_artifacts(path_df)
        log(f"Stored `{filename_df}`")
        mlflow_log_artifacts(path_figure)
        log(f"Stored `{filename_fig}`")


def _store_dict(x, name: str):
    """Store a dict in mlflow.

    Args:
        x ([type]): The dict to store
        name (str): The name of the dict
    """
    json_ = json.dumps(x, indent=4)
    with tempfile.TemporaryDirectory() as temporary_directory:
        filename_dict = name + ".json"
        path_dict = os.path.join(temporary_directory, filename_dict)
        f = open(path_dict, "w+")
        f.write(json_)
        f.close()
        mlflow_log_artifacts(path_dict)
        log(f"Stored `{filename_dict}`")


def _store_list(x, name: str):
    """Store a list in mlflow.

    Args:
        x ([type]): The list to store
        name (str): The name of the list
    """
    json_ = json.dumps(x, indent=4)
    with tempfile.TemporaryDirectory() as temporary_directory:
        filename_list = name + ".json"
        path_dict = os.path.join(temporary_directory, filename_list)
        f = open(path_dict, "w+")
        f.write(json_)
        f.close()
        mlflow_log_artifacts(path_dict)
        log(f"Stored `{filename_list}`")


def _store_metric(x, name: str):
    """Store a metric in mlflow.

    Args:
        x ([type]): The metric to store
        name (str): The name of the metric
    """
    mlflow.log_metric(name, x)


def _store_plot(x, name: str, show: bool = False):
    """Store a plot in MLFlow

    Args:
        x ([type]): The plot to store
        name (str): The name of the plot
        show (bool, optional): If true, also prints the plot to screen. Defaults to False.
    """
    with tempfile.TemporaryDirectory() as temporary_directory:
        filename_fig = name + ".jpeg"
        path_figure = os.path.join(temporary_directory, filename_fig)
        x.savefig(path_figure)
        mlflow_log_artifacts(path_figure)
        log(f"Stored `{filename_fig}`")
    if show:
        plt.show()


def store(x, name, show: bool = False):
    """Store x in mlflow.

    x can either be a dataframe, or a value.

    Args:
        x (): The thing to store
        name (str): The name of the thing
        show (bool, optional): If True, when storing a matplotlib.figure.Figure, prints the plot to screen.
                                Defaults to False.
    """
    if isinstance(x, pd.DataFrame):
        _store_dataframe(x, name)
    elif isinstance(x, matplotlib.figure.Figure):
        _store_plot(x, name, show)
    elif isinstance(x, int):
        _store_metric(x, name)
    elif isinstance(x, float):
        _store_metric(x, name)
    elif isinstance(x, dict):
        _store_dict(x, name)
    elif isinstance(x, list):
        _store_list(x, name)
    else:
        raise TypeError("%s type not implemented yet." % type(x))


def _plot_test_results_scatter(y_true, y_pred, show):
    """Scatter plot of the predicted vs true targets
    Args:
        y_true (pd.Series): The actual target values
        y_pred (pd.Series): The predicted target values
        show (bool, optional): If True, also show the charts on screen. Defaults to False.
    """
    import matplotlib.pyplot as plt

    fig, ax = plt.subplots()

    ax.plot(
        [y_true.min(), y_true.max()], [y_true.min(), y_true.max()], "r--", linewidth=2
    )

    ax.scatter(y_true, y_pred, alpha=0.3)

    ax.spines["top"].set_visible(False)
    ax.spines["right"].set_visible(False)

    ax.set_xlabel("Actual")
    ax.set_ylabel("Predicted")

    extra = plt.Rectangle(
        (0, 0), 0, 0, fc="w", fill=False, edgecolor="none", linewidth=0
    )

    store(fig, "Test results scatter", show)


def _plot_test_results_lines(y_true, y_pred, show):
    """Line plot of the predicted vs true targets
    Args:
        y_true (pd.Series): The actual target values
        y_pred (pd.Series): The predicted target values
        show (bool, optional): If True, also show the charts on screen. Defaults to False.
    """
    import matplotlib.pyplot as plt
    from matplotlib.dates import DateFormatter

    fig, ax = plt.subplots()

    ax.plot(y_true, "b-", y_pred, "r--")

    my_fmt = DateFormatter("%d")
    ax.xaxis.set_major_formatter(my_fmt)

    fig.autofmt_xdate()

    store(fig, "Test results lines", show)


def store_test_results(y_true: pd.Series, y_pred: pd.Series, show: bool = False):
    """Store the test results of a model

    Args:
        y_true (pd.Series): The actual target values
        y_pred (pd.Series): The predicted target values
        show (bool, optional): If True, also show the charts on screen. Defaults to False.
    """
    _plot_test_results_scatter(y_true, y_pred, show)
    _plot_test_results_lines(y_true, y_pred, show)


# def load_model(project_name: str = None, model_name: str = None, run_id: str = None):
#     """[summary]

#     @TODO: Load a specific model (eg. after inspecting the results in the dashboard)
#     @TODO: Limit to 'registered models' in MLFlow?

#     Args:
#         project_name (str, optional): [description]. Defaults to None.
#         model_name (str, optional): [description]. Defaults to None.
#         run_id (str, optional): [description]. Defaults to None.
#     """
#     pass


def store_model(model, extra_data: Dict = None):
    mlflow.keras.log_model(model, "model")
    json_ = json.dumps(extra_data)
    mlflow.set_tag("extra_data", json_)


def load_model(model_name: str, stage: str = None, version: int = None):
    import mlflow.pyfunc

    model = mlflow.pyfunc.load_model(model_uri=f"models:/{model_name}/{stage}")

    return model


def load_latest_model(project_name: str, model_name: str, local_mlflow: bool = False):
    """Load the latest model for a given project and model

    @TODO: Set a timeout (raise error if no response from MLFlow)

    Args:
        project_name (str): The name of the project
        model_name (str): The name of the model
        local_mlflow (bool, optional): If True, uses the local MLFlow. Defaults to False.
    """
    import mlflow.sklearn

    if local_mlflow:
        tracking_uri = pathlib.Path(
            os.path.join(os.path.expanduser("~"), "mlruns")
        ).as_uri()
    else:
        tracking_uri = config.mlflow.tracking_url
    mlflow.set_tracking_uri(tracking_uri)
    mlflow.set_experiment(project_name)
    df = mlflow.search_runs()
    # df = df[df["tags.mlflow.runName"] == model_name]
    df = df[df["tags.model_name"] == model_name]
    df = df[df["status"] == "FINISHED"]
    try:
        df = df[~df["metrics.training_score"].isna()]
    except KeyError:
        df = df[~df["tags.mlflow.log-model.history"].isna()]
    latest_run_id = df.iloc[0]["run_id"]
    run_res = mlflow.get_run(latest_run_id)

    model_path = os.path.join(run_res.info.artifact_uri.replace("file://", ""), "model")
    extra_data = json.loads(df.iloc[0]["tags.extra_data"])

    # TODO: FIX
    # try:
    #     model = mlflow.sklearn.load_model(model_path)
    # except mlflow.exceptions.MlflowException:
    model = mlflow.keras.load_model(model_path)
    model.extra_data = extra_data
    model.model_version = latest_run_id
    return model


def set_note(note: str, overwrite: bool = False) -> None:
    """Add a note to the current experiment.

    This will add a new note to the project, or overwrite an existing note
    if `overwrite=True`.

    Example:
        >>> set_note(
        >>>    "This is an inital test of the basic model for reservoir levels"
        >>> )

    Args:
        note (str): The text to add.
        overwrite (bool): Whether to overwrite an existing note. Defaults to False.
    """
    mlflow.set_tag("mlflow.note.content", note)


def add_tag(name: str, value: str) -> None:
    """Add a tag to the current experiment.

    All tags have a `name` AND a `value`, eg: "color" "blue".

    Example:
        >>> add_tag("reference", "https://arxiv.org/pdf/2010.99999.pdf")
        >>> # Adds one tag to the current experiment

    Args:
        name (str): The name of the tag to add
        value (str): The value of the tag to add
    """
    mlflow.set_tag(name, value)
